/*******************************************************/
/* Linux  GNU     C++ Compiler                         */
/* Name : tsa.cc                                       */
/* Autor: Harald Stauss                                */
/*******************************************************/

#include <limits.h>
#include "tsa.h"
 
/* ----------------------------- */
/* Implementation der Funktionen */
/* ----------------------------- */

REGIST::REGIST()		/* Constructor */
{
  n=0;
  ch=(CHANNEL *) malloc(sizeof(CHANNEL));
  comments=(COMMENTS *) malloc(sizeof(COMMENTS));
  comments->comment=(char **) malloc(sizeof(char *));
  comments->time=(unsigned long *) malloc(sizeof(unsigned long));
  comments->n=0;
}

REGIST::~REGIST()		/* Destructor */
{
   short        i;
   unsigned int j;
   if(n>0) for (i=0; i<n; i++) free(ch[i].data);
   free(ch);
   if (comments->n>0)
     for (j=0; j<comments->n; j++)
       free((char *)comments->comment[j]);
   free((char *)comments->comment);
   free((unsigned long *)comments->time);
   free(comments);
}

/* Funktion zum Lesen der Werte aus ASCII Format */
int REGIST::ReadASCII(char *fname, float sr, char delim,
		      unsigned long bc,
		      unsigned long ec,
		      unsigned long br,
		      unsigned long er)
{
  ifstream indat(fname);
  int      i,rc;
  short    ncol;
  char     *linie, *lstart, *clinie, *ilinie, *istart;
  char     cd[6];
  float    wert;
  unsigned long   ni,ndata,chl=100000;
  
  /* Initialisierung */
  cd[0]=delim; cd[1]=' '; cd[2]='\t'; cd[3]=(char)0x0D; cd[4]='\n'; cd[5]='\0';
  rc=0;
  ndata=0;
  linie=(char *) malloc(1024);
  lstart=linie;
  ilinie=(char *) malloc(1024);
  istart=ilinie;

  // validate opening of file;
  if (!indat) {
    cout << "Can't open file: " << fname << "\n";
    rc=-1;
    goto ende;
  }
    
  // Validate br and er;
  if (er==0) er=ULONG_MAX;
  if (br<1) {
    cout << "start value too low!\n";
    rc=-1;
    goto ende;
  }
  if (er<br) {
    cout << "start value larger than end value! \n";
    rc=-1;
    goto ende;
  }

  // Anzahl der Spalten feststellen;
  for(ni=1; ni<br; ni++) indat.getline(linie, 1023);
  indat.getline(linie,1023);
  ndata++;
  strcpy(ilinie,linie);
  clinie=strtok(ilinie,cd);
  ncol=1;
  while (clinie) {
    clinie=strtok(NULL, cd);
    ncol++;
  } /* endwhile */
  ncol--;
  
  /* bc and ec validate */
  if (bc<1){
    cout << "begin of columns to low!\n";
    goto ende;
  }
  if (ec==0) ec=ncol;
  if (ec<bc){
    cout << "begin column smaller than end column!\n";
    goto ende;
  }
  if (ec>ncol) {
    cout << "end column larger than number of columns in file!\n";
    goto ende;
  }

  /* first allocation of data array */
  ch=(CHANNEL *) realloc(ch,(n+ec-bc+1)*sizeof(CHANNEL));
  if (!ch) {
    cout << "allocation failure in ReadASCII \n";
    rc=-1;
    goto ende;
  }
  for(i=n; i<n+ec-bc+1; i++) {
    ch[i].data=(float *) malloc(chl*sizeof(float));
    if (!ch[i].data) {
      cout << "allocation failure in ReadASCII \n";
      rc=-1;
      goto ende;
    }
    ch[i].n=0;
    ch[i].height=0;
    ch[i].color=0;
  }
  strcpy(ilinie,linie);
  clinie=strtok(ilinie,cd);
  for (i=0; i<ncol; i++) {
    if (sscanf(clinie,"%f",&wert)!=-1 && i+1>=bc && i+1<=ec) {
      ch[n+i-bc+1].data[ch[n+i-bc+1].n] = wert;
      ch[n+i-bc+1].n++;
      ch[n+i-bc+1].min=wert;
      ch[n+i-bc+1].max=wert;
    }
    clinie=strtok(NULL,cd);
  }
  /* Daten einlesen */
  while (!indat.eof() && ndata<=er-br) {
    indat.getline(linie,1023);
    ndata++;
    strcpy(ilinie,linie);
    clinie=strtok(ilinie,cd);
    if (clinie) {
      for (i=0; i<ncol; i++) {
	if (sscanf(clinie,"%f",&wert)!=-1 && i+1>=bc && i+1<=ec) {
	  ch[n+i-bc+1].data[ch[n+i-bc+1].n] = wert;
	  ch[n+i-bc+1].n++;
	  if (ch[n+i-bc+1].min>wert) ch[n+i-bc+1].min=wert;
	  if (ch[n+i-bc+1].max<wert) ch[n+i-bc+1].max=wert;
	}
	clinie=strtok(NULL,cd);
	if (!clinie && i<ncol-1) {
	  cout << "data file corrupted at line: " << ndata << "\n";
	  rc=-1;
	  goto ende;
	}
	// realloc ?;
	if (fmod(ch[n+i-bc+1].n,chl)==0) {
	  ch[n+i-bc+1].data=(float *)realloc(ch[n+i-bc+1].data,
                                 ((ch[n+i-bc+1].n/chl)+1)*chl*sizeof(float));
	  if (!ch[n+i-bc+1].data) {
	    cout << "allocation failure! \n";
	    rc=-1;
	    goto ende;
	  }
	} // endif;
      } // endfor;
    } //endif;
  } // endwhile;
  for (i=n; i<n+ec-bc+1; i++) {
    ch[i].data=(float *)realloc(ch[i].data,ch[i].n*sizeof(float));
    ch[i].sr=sr;
    strcpy(ch[i].name,"?");
    strcpy(ch[i].unit,"?");
  } /* endfor */
  n+=ec-bc+1;

 ende:
  indat.close();
  ilinie=istart; linie=lstart;
  free(ilinie); free(linie);
  return(rc);
}

/*Funktion zum Schreiben der Werte auf ASCII Format */
int REGIST::WriteASCII(char *fname, char delim,
		       unsigned long bc,
		       unsigned long ec,
		       unsigned long br,
		       unsigned long er)
{
  int           i,rc=0;
  float         wert;
  ofstream outdat(fname);
  unsigned long ni,lch;
  
  // validate file
    if (!outdat) {
      cout << "Cannot open file! \n";
      rc=-1;
      goto ende;
    }
  
  // validate columns;
  if (ec==0) ec=n;
  if (bc<1 || ec>n || ec<bc) {
    cout << "column specification fault!\n";
    rc=-1;
    outdat.close();
    goto ende;
  }
  
  // find longest channel;
  lch=0;
  for (i=0; i<n; i++) if (lch<ch[i].n) lch=ch[i].n;
  
  // validate rows;
  if (er==0) er=lch;
  if (br<1 || er>lch || er<br) {
    cout << "row specification fault!\n";
    rc=-1;
    outdat.close();
    goto ende;
  }

  // write data;
  for(ni=br-1; ni<er; ni++) {
    for(i=bc-1; i<ec; i++) {
      if(ni<ch[i].n) {
	outdat << ch[i].data[ni];
	if (i<ec-1) outdat << delim;
      }
      else {
	outdat << " ";
	if (i<ec-1) outdat << delim;
      }
    }
    outdat << "\n";
  }

  outdat.close();
  
 ende:
  return(rc);
}

// function to write TSA format;
int REGIST::WriteTSA(char *fname)
{
  int i,rc=0;
  unsigned long ni;
  ofstream outdat(fname);

  // validate outstream;
  if (!outdat) {
    cout << "cant open file for output!\n";
    rc=-1;
    goto ende;
  }

  // write header as text;
  outdat << n << "\n";
  outdat << "name\t" << "unit\t" << "sr\t" << "n\n";
  for (i=0; i<n; i++)  outdat << ch[i].name << "\t" << ch[i].unit << "\t"
      << ch[i].sr << "\t" << ch[i].n << "\n";
  outdat.close();

  // write data as binary;
  outdat.open(fname, ios::bin | ios::app,0);
  for (i=0; i<n; i++) 
    for (ni=0; ni<ch[i].n; ni++)
      outdat.write((unsigned char *) &ch[i].data[ni], sizeof(float));
  outdat.close();

 ende:
  return(rc);
}

// Function to read TSA format;
int REGIST::ReadTSA(char *fname)
{
  short nch;
  int   rc=0,i;
  unsigned long ni;
  char  linie[1024];
  char  *clinie;
  char  z;
  ifstream indat(fname);

  // validate indat;
  if (!indat) {
    rc=-1;
    cout << "cant open file for input \n";
    goto ende;
  }

  // header einlesen;
  indat.getline(linie,1023);
  nch=atoi(linie);
  indat.getline(linie,1023);
  // allocation of memory for new channels;
  ch=(CHANNEL *)realloc(ch,(n+nch)*sizeof(CHANNEL));
  if (!ch) {
    cout << "allocation failure in ReadTSA \n";
    rc=-1;
    goto ende;
  }
  for (i=0; i<nch; i++) {
    indat.getline(linie,1023);
    clinie=strtok(linie,"\t\n");
    strcpy(ch[i+n].name,clinie);
    clinie=strtok(NULL,"\t\n");
    strcpy(ch[i+n].unit,clinie);
    clinie=strtok(NULL,"\t\n");
    ch[i+n].sr=atof(clinie);
    clinie=strtok(NULL,"\t\n");
    ch[i+n].n=strtoul(clinie,NULL,10);
    ch[i+n].data=(float *) malloc(ch[i+n].n*sizeof(float));
    if (!ch[i+n].data) {
      cout << "allocation failure in ReadTSA \n";
      rc=-1;
      goto ende;
    }
    ch[i+n].color=0;
    ch[i+n].height=0;
  }
  indat.close();
  // open file as binary and input data;
  indat.open(fname, ios::in | ios::bin, 0);
  if (!indat) {
    rc=-1;
    cout << "could not open file!\n";
    goto ende;
  }

  // search for begin of data;
  i=-2; // there are 2 more lines than channels in the header;
  while(i<nch) {
    indat.get(z);
    if (z=='\n') i++;
  }

  for (i=n; i<n+nch; i++) {
    ch[i].min=10e32;
    ch[i].max=-10e32;
    for (ni=0; ni<ch[i].n; ni++) {
      ch[i].data[ni]=0;
      indat.read((unsigned char *) &ch[i].data[ni], sizeof(float));
      if (ch[i].data[ni]<ch[i].min) ch[i].min=ch[i].data[ni];
      if (ch[i].data[ni]>ch[i].max) ch[i].max=ch[i].data[ni];
    }
  }
  indat.close();
  n+=nch;

 ende:
  return(rc);
}

// Function to read MAD format;
int REGIST::ReadMAD(char *fname)
{
  float version, sr;
  short nch, c_exist;
  int   rc=0, i;
  unsigned int  j, ci, cj;
  unsigned long ni, ncomments;
  char  linie[1024], blanck[2];
  char  *clinie, *cptime, *cpsubtime, *cpcomment;
  char  z;
  ifstream indat(fname);

  strcpy(blanck, " ");
  // validate indat;
  if (!indat) {
    rc=-1;
    cout << "cant open file for input \n";
    goto ende;
  }
  // test if MAD file;
  indat.getline(linie,1023);
  if (strncmp(linie,"MAD", 3)) {
    cout << "This is not a MAD file \n";
    rc=-1;
    goto ende;
  }
  // get version and number of channels;
  clinie=linie+3;
  sscanf(clinie,"%f %d",&version, &nch);
  // allocation of memory for new channels;
  ch=(CHANNEL *)realloc(ch,(n+nch)*sizeof(CHANNEL));
  if (!ch) {
    cout << "allocation failure in ReadMAD \n";
    rc=-1;
    goto ende;
  }
  // header einlesen;
  indat.getline(linie,1023); // headline only comments;
  for (i=0; i<nch; i++) {
    strcpy(ch[i+n].name,"?");
    strcpy(ch[i+n].unit,"?");
    indat.getline(linie,1023);
    clinie=strtok(linie," \t\n");
    ch[i+n].n=strtoul(clinie,NULL,10);
    clinie=strtok(NULL," \t\n");
    ch[i+n].sr=atof(clinie);
    clinie=strtok(NULL," \t\n");
    ch[i+n].color=atoi(clinie);
    clinie=strtok(NULL," \t\n");
    ch[i+n].height=atoi(clinie);
    ch[i+n].data=(float *) malloc(ch[i+n].n*sizeof(float));
    if (!ch[i+n].data) {
      cout << "allocation failure in ReadMAD \n";
      rc=-1;
      goto ende;
    }
  }
  // read comments;
  indat.getline(linie,1023);
  while (strncmp(linie,"data",4)) {
    /* extract time (sec) and comment */
    cptime=strtok(linie,"\t");
    cpcomment=strtok(NULL, "\n");
    if (!cpcomment) cpcomment=blanck;
    cpsubtime=strtok(cptime,":");
    ni=60*60*strtoul(cpsubtime, NULL, 10);
    cpsubtime=strtok(NULL, ":");
    ni+=60*strtoul(cpsubtime, NULL, 10);
    cpsubtime=strtok(NULL, "\n");
    ni+=strtoul(cpsubtime, NULL, 10);
    /* check if time and comment already exist */
    c_exist=0;
    for (j=0; j<comments->n; j++) 
      if ((ni==comments->time[j]) && !strcmp(cpcomment, comments->comment[j])) c_exist=1;
    if (!c_exist) {
      comments->time=(unsigned long *)realloc(comments->time, (comments->n+1)*sizeof(unsigned long));
      comments->comment=(char **)realloc(comments->comment, (comments->n+1)*sizeof(char *));
      comments->comment[comments->n]=(char *)malloc(strlen(cpcomment)+1);
      strcpy(comments->comment[comments->n], cpcomment);
      comments->time[comments->n]=ni;
      comments->n++;
    }
    indat.getline(linie,1023);
  }
  /* sort comments */
  if (comments->n>0) {
    for (cj=(comments->n-1); cj>0; cj--)
      for (ci=0; ci<cj; ci++) {
	if (comments->time[ci]>comments->time[ci+1]) {
	  /* exchange */
	  ni=comments->time[ci];
	  comments->time[ci]=comments->time[ci+1];
	  comments->time[ci+1]=ni;
	  cpcomment=comments->comment[ci];
	  comments->comment[ci]=comments->comment[ci+1];
	  comments->comment[ci+1]=cpcomment;
	}
      }
  }
  /* close file that was opened as text file */
  indat.close();
  // open file as binary and input data;
  indat.open(fname, ios::in | ios::bin, 0);
  if (!indat) {
    rc=-1;
    cout << "could not open file!\n";
    goto ende;
  }
  // search for begin of data;
  indat.getline(linie,1023);
  while (strncmp(linie,"data",4)) {
    indat.getline(linie,1023);
  }
  for (i=n; i<n+nch; i++) {
    ch[i].min=10e32;
    ch[i].max=-10e32;
    for (ni=0; ni<ch[i].n; ni++) {
      ch[i].data[ni]=0;
      indat.read((unsigned char *) &ch[i].data[ni], sizeof(float));
      if (ch[i].data[ni]<ch[i].min) ch[i].min=ch[i].data[ni];
      if (ch[i].data[ni]>ch[i].max) ch[i].max=ch[i].data[ni];
    }
  }
  indat.close();
  n+=nch;
 ende:
  return(rc);
}

// function to read WAV files;
int REGIST::ReadWAV(char *fname) {
  ifstream indat;
  int      rc=0, fl, sr, bps, i;
  char     riff[5];
  short    lsc, nch, bytesps;
  unsigned long ndata, ni;
  signed short  wert;

  // validate indat;
  indat.open(fname, ios::in | ios::bin, 0);
  if (!indat) {
    rc=-1;
    cout << "cant open file for input \n";
    goto ende;
  }
  // test if WAV file;
  indat.read((char *)riff, 4); riff[4]='\0';
  if (strncmp(riff,"RIFF", 4)) {
    cout << "This is not a WAV file \n";
    rc=-1;
    goto ende;
  }
  // get file length;
  indat.read((int *)&fl, 4);
  // get chunk type und sub chunk;
  indat.read((char *)riff, 4); riff[4]='\0';
  indat.read((char *)riff, 4); riff[4]='\0';
  // get length of sub chunk und format;
  indat.read((short *)&lsc, 2);
  indat.read((int *)&fl, 4);
  // get number of channels (MODUS);
  indat.read((short *)&nch, 2);
  // get sampling rate;
  indat.read((int *)&sr, 4);
  // get bytes per second;
  indat.read((int *)&bps, 4);
  // get bytes per sample;
  indat.read((short *)&bytesps, 2);
  // get bits per sample;
  indat.read((short *)&lsc, 2);
  // get Daten Chunk;
  indat.read((char *)&riff, 4); riff[4]='\0';
  // get length of data;
  indat.read((int *)&fl, 4);
  // got header; 

  // now read the data;
  ndata=fl/bytesps;
  // allocation of memory for new channels;
  ch=(CHANNEL *)realloc(ch,(n+nch)*sizeof(CHANNEL));
  if (!ch) {
    cout << "allocation failure in ReadWAV \n";
    rc=-1;
    goto ende;
  }
  for (i=0; i<nch; i++) {
    strcpy(ch[i+n].name,"sound");
    strcpy(ch[i+n].unit,"dB");
    ch[i+n].n=ndata;
    ch[i+n].sr=sr;
    ch[i+n].color=1;
    ch[i+n].height=0;
    ch[i+n].min=10e32;
    ch[i+n].max=-10e32;
    ch[i+n].data=(float *) malloc(ch[i+n].n*sizeof(float));
    if (!ch[i+n].data) {
      cout << "allocation failure in ReadWAV \n";
      rc=-1;
      goto ende;
    }
  }
  for (ni=0; ni<ndata; ni++) {
    for (i=n; i<n+nch; i++) {
      wert=0;
      indat.read((signed short *) &wert, bytesps/nch);
      ch[i].data[ni]=(float)wert;
      if (ch[i].data[ni]<ch[i].min) ch[i].min=ch[i].data[ni];
      if (ch[i].data[ni]>ch[i].max) ch[i].max=ch[i].data[ni];
    }
  }
  indat.close();
  n+=nch;

 ende:
  return(rc);
}



// function to add new channel
int REGIST::AddChannel(unsigned long ndata,
		       float sr=1, float min=0, float max=1,
		       int height=0, int color=0,
		       char *name="channel", char *unit="unit") {
  int rc=0;
  
  n++;
  ch=(CHANNEL *) realloc(ch,n*sizeof(CHANNEL));
  ch[n-1].data=(float *) malloc(ndata*sizeof(float));
  ch[n-1].n=ndata;
  ch[n-1].sr=sr;
  ch[n-1].min=min;
  ch[n-1].max=max;
  ch[n-1].height=height;
  ch[n-1].color=color;
  strcpy(ch[n-1].name, name);
  strcpy(ch[n-1].unit, unit);
  return rc;
}


// function to delete channel;
 int REGIST::DelChannel(short nch) {
  int           rc=0;
  short         i;
  unsigned long ni;

  if ((nch>=n) || (nch<0))  {
    rc=-1;
    goto ende;
  }
  
  for (i=nch+1; i<n; i++) {
    ch[i-1].data=(float *)realloc(ch[i-1].data, ch[i].n * sizeof(float));
    for (ni=0; ni<ch[i].n; ni++) ch[i-1].data[ni]=ch[i].data[ni];
    ch[i-1].n=ch[i].n;
    ch[i-1].sr=ch[i].sr;
    ch[i-1].min=ch[i].min;
    ch[i-1].max=ch[i].max;
    strcpy(ch[i-1].name, ch[i].name);
    strcpy(ch[i-1].unit, ch[i].unit);
  }

  n--;
  free(ch[n].data);
  ch=(CHANNEL *) realloc(ch, n*sizeof(CHANNEL));

  /* delete comments if no channels are left */
  if(n==0) {
    if (comments->n>0)
      for (ni=0; ni<comments->n; ni++) free((char *)comments->comment[ni]);
    comments->n=0;
  }

 ende:
  return rc;
}

